Coverage for F:\DEV\EMIZ\emiz\sltp.py: 16%

# Coverage for F:\DEV\EMIZ\emiz\sltp.py : 16%

## 250 statements 41 run 209 missing 0 excluded

Hot-keys on this page

r m x p toggle line displays

j k next/prev highlighted chunk

0 (zero) top of page

1 (one) first highlighted chunk

1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 32 33 34 35 36 37 38 39 40 41 42 43 44 45 46 47 48 49 50 51 52 53 54 55 56 57 58 59 60 61 62 63 64 65 66 67 68 69 70 71 72 73 74 75 76 77 78 79 80 81 82 83 84 85 86 87 88 89 90 91 92 93 94 95 96 97 98 99 100 101 102 103 104 105 106 107 108 109 110 111 112 113 114 115 116 117 118 119 120 121 122 123 124 125 126 127 128 129 130 131 132 133 134 135 136 137 138 139 140 141 142 143 144 145 146 147 148 149 150 151 152 153 154 155 156 157 158 159 160 161 162 163 164 165 166 167 168 169 170 171 172 173 174 175 176 177 178 179 180 181 182 183 184 185 186 187 188 189 190 191 192 193 194 195 196 197 198 199 200 201 202 203 204 205 206 207 208 209 210 211 212 213 214 215 216 217 218 219 220 221 222 223 224 225 226 227 228 229 230 231 232 233 234 235 236 237 238 239 240 241 242 243 244 245 246 247 248 249 250 251 252 253 254 255 256 257 258 259 260 261 262 263 264 265 266 267 268 269 270 271 272 273 274 275 276 277 278 279 280 281 282 283 284 285 286 287 288 289 290 291 292 293 294 295 296 297 298 299 300 301 302 303 304 305 306 307 308 309 310 311 312 313 314 315 316 317 318 319 320 321 322 323 324 325 326 327 328 329 330 331 332 333 334 335 336 337 338 339 340 341 342 343 344 345 346 347 348 349 | # coding=utf-8 """Simple Lua Python Parser""" import re import mpmath from natsort import natsorted import logging LOGGER = logging.getLogger('EMIZ').getChild(__name__) # noinspection SpellCheckingInspection ERRORS = { 'unexp_type_str': 'decoding error: string expected', 'unexp_end_string': 'Unexpected end of string while parsing Lua string.', 'unexp_end_table': 'Unexpected end of table while parsing Lua string.', 'mfnumber_minus': 'Malformed number (no digits after initial minus).', 'mfnumber_dec_point': 'Malformed number (no digits after decimal point).', 'mfnumber_sci': 'Malformed number (bad scientific format).', } class BaseSLTPError(Exception): """Base exception for SLTP module""" def __init__(self, *args): super().__init__(*args) class SLTPParsingError(BaseSLTPError): """Error during parsing""" def __init__(self, *args): super().__init__(*args) class SLTPEmptyObjectError(BaseSLTPError): """Error during parsing""" def __init__(self, *args): super().__init__(*args) class SLTP: """Simple Lua Python Parser""" def __init__(self): LOGGER.debug('instantiating parser') self.text = '' self.ch = '' self.at = 0 self.len = 0 self.depth = 0 self.qual = None self.space = re.compile(r'\s', re.M) self.alnum = re.compile(r'\w', re.M) self.line_end = re.compile(r'^(?P<intro>[ \t]*\})(?P<comment> -- end of \[.*\]),$') self.newline = '\n' self.tab = '\t' self.tab = ' ' def decode(self, text): """Decode a Lua string to an dictionary :type text: str :rtype: dict :param text: string to decode :return: dictionary """ LOGGER.debug('decoding text to dictionary') if not text or type(text) is not str: raise SLTPParsingError(ERRORS['unexp_type_str']) LOGGER.debug('extracting qualifier') qual = re.compile(r'^(?P<value>(dictionary|mission|mapResource) = ?)\n') match = qual.match(text) if match is None: raise ValueError('qualifier not found; first line: {}'.format(text.split('\n')[0])) self.qual = match.group('value') text = qual.sub('', text) reg = re.compile(r' -- .*[^\\]$', re.M) text = reg.sub('', text) self.text = text self.at, self.ch, self.depth = 0, '', 0 self.len = len(text) self.next_chr() result = self.value() return result, self.qual def encode(self, obj, qualifier: str): """Encodes a dictionary-like object to a Lua string :param qualifier: :param obj: object to encode :return: valid Lua string """ LOGGER.debug('encoding dictionary to text') if not obj: if qualifier.replace('=', '').rstrip() == 'mapResource': # Accept empty mapResource return '{}\n{{\n}} -- end of {}\n'.format(qualifier, qualifier.replace('=', '').rstrip()) else: LOGGER.error('{}\n{{\n}} -- end of {}\n'.format(qualifier, qualifier.replace('=', '').rstrip())) raise SLTPEmptyObjectError(qualifier) # return '{}\n{{\n}} -- end of {}\n'.format(qualifier, qualifier.replace('=', '').rstrip()) self.depth = 0 out = [] s = self.__encode(obj) lines = s.split(self.newline) for line in lines: m = self.line_end.match(line) if m: out.append('{},{}'.format(m.group('intro'), m.group('comment'))) else: out.append(line) return '{}{} -- end of {}\n'.format(qualifier, self.newline.join(out), qualifier.replace('=', '').rstrip()) def __encode(self, obj, dict_name=None): # noqa C901 s = '' tab = self.tab newline = self.newline if isinstance(obj, str): s += '"%s"' % obj.replace('"', '\\"') elif isinstance(obj, bool): s += str(obj).lower() elif isinstance(obj, (int, float, complex, mpmath.mpf)): s += str(obj) elif isinstance(obj, (list, tuple, dict)): # Ladies and gentlemen, please take a minute to behold the following code. # Breathe slowly, let it sink in for a minute. # Yes, I agree, this sucks *so* much I should hang for it. # Now, deal with it ^^ self.depth += 1 # noinspection PyTypeChecker if not isinstance(obj, dict) and len(filter( lambda x: isinstance(x, (int, float, mpmath.mpf)) or (isinstance(x, str) and len(x) < 10), obj )) == len(obj): newline = tab = '' dp = tab * self.depth s += '%s%s{%s' % (newline, tab * (self.depth - 1), newline if len(obj) > 0 else '') if isinstance(obj, dict): s += (',%s' % newline).join( [dp + '[{}] ={}{}'.format(k, '' if isinstance(obj[k], (list, tuple, dict)) else ' ', self.__encode(obj[k], k)) if type( k) is int else dp + '["{}"] ={}{}'.format(k, '' if isinstance(obj[k], (list, tuple, dict)) else ' ', self.__encode(obj[k], k)) for k in natsorted(obj.keys(), key=str)]) else: s += (',%s' % newline).join([dp + self.__encode(el) for el in obj]) self.depth -= 1 if len(obj) > 0: s += ',' s += '%s%s}' % (newline, tab * self.depth) if isinstance(obj, dict): try: int(dict_name) s += ' -- end of [{}]'.format(dict_name) except (ValueError, TypeError): if dict_name is not None: s += ' -- end of ["{}"]'.format(dict_name) return s # noinspection PyMissingOrEmptyDocstring def white(self): while self.ch: if self.space.match(self.ch): self.next_chr() else: break # noinspection PyMissingOrEmptyDocstring def next_chr(self): if self.at >= self.len: self.ch = None return None self.ch = self.text[self.at] self.at += 1 return True # noinspection PyMissingOrEmptyDocstring def value(self): self.white() if not self.ch: return if self.ch == '{': o = self.object() ret = dict() for k in natsorted(o.keys()): ret[k] = o[k] return ret if self.ch == '[': self.next_chr() if self.ch in ['"', "'", '[']: return self.string(self.ch) if self.ch.isdigit() or self.ch == '-': return self.number() return self.word() # noinspection PyMissingOrEmptyDocstring def string(self, end=None): s = '' start = self.ch if end == '[': end = ']' if start in ['"', "'", '[']: while self.next_chr(): if self.ch == end: self.next_chr() if start != '[' or self.ch == ']': return s if self.ch == '\\' and start == end: self.next_chr() if self.ch != end: s += '\\' s += self.ch raise SLTPParsingError(ERRORS['unexp_end_string']) # noinspection PyMissingOrEmptyDocstring def object(self): # noqa C901 o = dict() k = '' idx = 0 numeric_keys = False self.depth += 1 self.next_chr() self.white() if self.ch and self.ch == '}': self.depth -= 1 self.next_chr() return o # Exit here else: while self.ch: self.white() if self.ch == '{': o[idx] = self.object() idx += 1 continue elif self.ch == '}': self.depth -= 1 self.next_chr() if k: o[idx] = k if not numeric_keys and len( [ key for key in o if type(key) in (str, float, bool, tuple, mpmath.mpf) ] ) == 0: ar = [] for key in o: ar.insert(key, o[key]) o = ar return o # or here else: if self.ch == ',': self.next_chr() continue else: k = self.value() if self.ch == ']': numeric_keys = True self.next_chr() self.white() if self.ch == '=': self.next_chr() self.white() o[k] = self.value() idx += 1 k = '' elif self.ch == ',': self.next_chr() self.white() o[idx] = k idx += 1 k = '' raise SLTPParsingError(ERRORS['unexp_end_table']) # Bad exit here # noinspection PyMissingOrEmptyDocstring def word(self): s = '' if self.ch != '\n': s = self.ch while self.next_chr(): if self.alnum.match(self.ch): s += self.ch else: if re.match('^true$', s, re.I): return True elif re.match('^false$', s, re.I): return False elif s == 'nil': return None return str(s) # noinspection PyMissingOrEmptyDocstring def number(self): # noqa C901 # noinspection PyMissingOrEmptyDocstring def next_digit(err): _n = self.ch self.next_chr() if not self.ch or not self.ch.isdigit(): raise SLTPParsingError(err) return _n n = '' if self.ch == '-': n += next_digit(ERRORS['mfnumber_minus']) n += self.digit() if n == '0' and self.ch in ['x', 'X']: n += self.ch self.next_chr() n += self.hex() else: if self.ch and self.ch == '.': n += next_digit(ERRORS['mfnumber_dec_point']) n += self.digit() if self.ch and self.ch in ['e', 'E']: n += self.ch self.next_chr() if not self.ch or self.ch not in ('+', '-'): raise SLTPParsingError(ERRORS['mfnumber_sci']) n += next_digit(ERRORS['mfnumber_sci']) n += self.digit() try: return int(n, 0) except ValueError: return mpmath.mpf(n) # noinspection PyMissingOrEmptyDocstring def digit(self): n = '' while self.ch and self.ch.isdigit(): n += self.ch self.next_chr() return n # noinspection PyMissingOrEmptyDocstring def hex(self): n = '' while self.ch and (self.ch in 'ABCDEFabcdef' or self.ch.isdigit()): n += self.ch self.next_chr() return n sltp = SLTP()